import numpy as np
from sklearn.metrics import auc
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class AUCP(BaseThresholder):
    """AUCP class for Area Under Curve Precentage thresholder.

       Use the area under the curve to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond where the auc of the kde is less
       than the (median + abs(median-mean)) percent of the total kde auc
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        pass

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.

        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Generate KDE
        val, dat_range = gen_kde(decision,0,1,len(decision)*2)
        val = normalize(val)

        # Get the total area under the curve
        tot_area = auc(dat_range,val)

        # Get area percentage limit
        med = np.median(decision)
        perc = med+abs(med-np.mean(decision))
        

        # Apply the limit to where the area is less than that limit percentage
        # of the total area under the curve
        limit = 1
        for i in range(len(dat_range)):
   
            splt_area = auc(dat_range[i:], val[i:])

            if splt_area<perc*tot_area:
                limit = dat_range[i]
                break

        self.thresh_ = limit
        
        return cut(decision, limit)



import numpy as np
import scipy.stats as stats
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut

# https://github.com/bhattbhavesh91/outlier-detection-grubbs-test-and-generalized-esd-test-python/blob/master/generalized-esd-test-for-outliers.ipynb

def grubbs_stat(y):
    
    dev = np.abs(y - y.mean())
    max_ind = np.argmax(dev)
    Gcal = np.max(dev)/ y.std()

    return Gcal, max_ind

def calc_crit(size, alpha):
    
    t_dist = stats.t.ppf(1 - alpha / (2 * size), size - 2)
    numerator = (size - 1) * np.sqrt(np.square(t_dist))
    denominator = np.sqrt(size) * np.sqrt(size - 2 + np.square(t_dist))
    critical_value = numerator / denominator

    return critical_value


class GESD(BaseThresholder):
    """GESD class for Generalized Extreme Studentized Deviate thresholder.

       Use the generalized extreme studentized deviate to evaluate a
       non-parametric means to threshold scores generated by the decision_scores
       where outliers are set to any less than the smallest detected outlier
       
       Paramaters
       ----------

       max_outliers : int, optional (default=None)
           if None max_outliers will be set to half the size of the dataset

       alpha : float, optional (default=0.05)
           significance level

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, max_outliers=None, alpha=0.05):

        self.max_outliers = max_outliers
        self.alpha = alpha


    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        arr = decision.copy()

        limit = 1.1

        if self.max_outliers is None:
            self.max_outliers = int(len(decision)/2)

        for i in range(self.max_outliers):
            Gc = calc_crit(len(arr), self.alpha)
            Gs, max_index = grubbs_stat(arr)
            
            if (Gs>Gc) & (arr[max_index]<limit):
                limit = arr[max_index]
            arr = np.delete(arr, max_index)

        self.thresh_ = limit

        return cut(decision, limit)



import numpy as np
import scipy.signal as signal
from scipy import integrate
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde

#https://github.com/geomdata/gda-public/blob/master/timeseries/curve_geometry.pyx

def mollifier(time, position, refinement=5, width=1.0):

    N = len(position)

    delta = (time[-1]-time[0])/(N-1)

    ### compute boundary space padding
    left_pad = np.arange(time[0],time[0]-(width+delta),step=-delta)
    left_pad = np.flipud(left_pad)[:-1]
    left_pad_num = left_pad.shape[0]
    right_pad = np.arange(time[-1],time[-1]+(width+delta),step=delta)[1:]
    right_pad_num = right_pad.shape[0]
    time_pad = np.concatenate((left_pad,time,right_pad))

    ### compute boundary score padding
    position_pad = np.pad(position,(left_pad_num,right_pad_num),'edge')

    ### Define a new smaller space scale s, ds (here we a evenly spaced)
    s, ds = np.linspace(time_pad[0],time_pad[-1],
                           (refinement)*time_pad.shape[0],
                           retstep=True)
    right_pad_num = (refinement)*right_pad_num
    left_pad_num = (refinement)*left_pad_num
    position_interp = np.interp(s,time_pad,position_pad)

    ### Compute the mollifier kernel
    norm_const,err = integrate.quad(lambda x: np.exp(1.0/(x**2-1.0)),-1.0,1.0)
    norm_const = 1.0/norm_const

    ### Compute the mollifier rho
    p = np.abs((s - (s[0]+s[-1])/2.0)/width)
    r = np.zeros_like(s)
    q = p[p<1.0]
    r[p<1.0] = np.exp(1.0/(q**2-1.0))
    rho = (norm_const/width)*r

    ### Perform convolution to make smooth reconstruction
    if s.shape[0] > 500:
        #print('Convolving with FFT.')
        smooth = signal.fftconvolve(ds*position_interp,rho,mode='same')
    else:
        #print('Convolving without FFT.')
        smooth = np.convolve(ds*position_interp,rho,mode='same')

    ### remove padding
    s = s[left_pad_num:-right_pad_num]
    smooth = smooth[left_pad_num:-(right_pad_num)]

    return np.asarray(smooth)


class MOLL(BaseThresholder):
    """MOLL class for Friedrichs' mollifier thresholder.

       Use the Friedrichs' mollifier to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond one minus the  maximum of the smoothed
       dataset via convolution
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        pass

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)
        dat_range = np.linspace(0,1,len(decision))

        decision = normalize(decision)

        # Set the inliers to be where the 1-max(smoothed scores)
        limit = 1-np.max(mollifier(dat_range, np.sort(decision)))

        self.thresh_ = limit

        return cut(decision, limit)



from __future__ import division
from __future__ import print_function

import warnings

import abc
import six

@six.add_metaclass(abc.ABCMeta)
class BaseThresholder(object):
    """Abstract class for all outlier detection thresholding algorithms.

    .. warning::
    pyod would stop supporting Python 2 in the future. Consider move to
    Python 3.5+.

       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    @abc.abstractmethod
    def __init__(self):

        self.thresh_ = None

    @abc.abstractmethod
    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.

        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        pass




import numpy as np
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut


class IQR(BaseThresholder):
    """IQR class for Inter-Qaurtile Region thresholder.

       Use the inter-quartile region to evaluate a non-parametric
       means to threshold scores generated by the decision_scores
       where outliers are set to any value beyond the third quartile
       plus 1.5 times the inter-quartile region

       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(IQR, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        try:
            # First quartile (Q1)
            P1 = np.percentile(decision, 25, interpolation='midpoint')
  
            # Third quartile (Q3)
            P3 = np.percentile(decision, 75, interpolation='midpoint')
        except TypeError:
            # First quartile (Q1)
            P1 = np.percentile(decision, 25, method='midpoint')
  
            # Third quartile (Q3)
            P3 = np.percentile(decision, 75, method='midpoint')

        # Calculate IQR and generate limit
        iqr = abs(P3-P1)
        limit = P3 + 1.5*iqr

        self.thresh_ = limit

        return cut(decision, limit)




import numpy as np
from scipy.cluster.vq import kmeans
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class KMEANS(BaseThresholder):
    """KMEANS class for KMEANS clustering thresholder.

       Use the kmeans clustering to evaluate a non-parametric means to
       threshold scores generated by the decision_scores where outliers
       are set to any value beyond the radius distance of the single
       cluster centroid

       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(KMEANS, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Find the most prominent kmeans center and radius
        codebook, distortion = kmeans(decision, 1)

        # Set limit to the radius distance from the center
        limit = codebook + distortion

        self.thresh_ = limit

        return cut(decision, limit)




import numpy as np
import scipy.stats as stats
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class ZSCORE(BaseThresholder):
    """ZSCORE class for ZSCORE thresholder.

       Use the zscore to evaluate a non-parametric means to threshold
       scores generated by the decision_scores where outliers are set
       to any value beyond a zscore of one
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(ZSCORE, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Get the zscore of the decision scores
        zscore = np.abs(stats.zscore(decision))

        # Set the limit to where the zscore is 1
        scores = np.zeros(len(decision), dtype=int)
        mask = np.where(zscore>=1.0)
        scores[mask] = 1

        self.thresh_ = None

        return scores



import numpy as np
import scipy.stats as stats
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut


class QMCD(BaseThresholder):
    """QMCD class for Quasi-Monte Carlo Discreprancy thresholder.

       Use the quasi-Monte Carlo discreprancy to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond and percentile or quantile of one minus the
       descreperancy (Note** A discrepancy quantifies the distance between the
       continuous uniform distribution on a hypercube and the discrete uniform
       distribution on distinct sample points)
       
       Paramaters
       ----------

       method : str, optional (default='WD')
        {'CD', 'WD', 'MD', 'L2-star'}
        CD:      Centered Discrepancy
        WD:      Wrap-around Discrepancy
        MD:      Mix between CD/WD
        L2-star: L2-star discrepancy

       lim : str, optional (default='P')
        {'Q', 'P'}
        Q: Use quntile limiting
        P: Use percentile limiting
        

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, method='WD', lim='P'):

        super(QMCD, self).__init__()
        self.method = method
        self.lim = lim

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Get the quasi Monte-Carlo discrepancy of the labels
        disc = stats.qmc.discrepancy(decision.reshape(-1,1), method=self.method)

        # Set the limit to either the quantile or percentile of 1-discrepancy
        if self.lim=='Q':
            limit = np.quantile(decision, 1.0-disc)
        elif self.lim=='P':
            try:
                limit = np.percentile(decision, (1.0-disc)*100, method='midpoint')
            except TypeError:
                limit = np.percentile(decision, (1.0-disc)*100, interpolation='midpoint')

        self.thresh_ = limit

        return cut(decision, limit)



import numpy as np
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize


class CLF(BaseThresholder):
    """CLF class for Trained Classifier thresholder.

       Use the trained linear classifier to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond 0
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        self.m = 6.252101632002534
        self.c = -2.7630070467775503

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Calculate expected y
        pred = self.m*decision + self.c

        # Determine labels
        pred[pred>0] = 1
        pred[pred<=0] = 0

        self.thresh_ = None

        return pred.astype(int)



import numpy as np
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class FGD(BaseThresholder):
    """FGD class for Fixed Gradient Descent thresholder.

       Use the fixed gradient descent to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond where the first derivative of the kde
       with respect to the decision scores passes the second inflection point
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        pass

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Generate KDE
        val, dat_range = gen_kde(decision,0,1,len(decision)*3)

        # Calculate the first derivative of the KDE with respect
        # to the data range
        deriv = np.gradient(val, dat_range[1]-dat_range[0])

        count = 0    
        ind = 0

        # Set the limit where the first derivative curve passes through the
        # origin a second time. (the first time occurs at the kde max peak)
        for i in range(len(deriv)-1):

            if (deriv[i]>0)&(deriv[i+1]<=0):
                count+=1
                if count==2:
                    ind = i
                    break

        limit = dat_range[ind]
        self.thresh_ = limit

        return cut(decision, limit)



import numpy as np
import scipy.stats as stats


def normalize(data):

    return ((data - data.min()) / (data.max() - data.min()))


def cut(decision, limit):

    labels = np.zeros(len(decision), dtype=int)

    labels[decision>=limit] = 1

    return labels


def gen_kde(data, lower, upper, size):

    # Create a KDE of the labels
    kde = stats.gaussian_kde(data)
    dat_range = np.linspace(lower,upper,size)

    return kde(dat_range), dat_range



import numpy as np
import scipy.stats as stats
from scipy.special import erfc 
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut


class CHAU(BaseThresholder):
    """CHAU class for Chauvenet's criterion thresholder.

       Use the Chauvenet's criterion to evaluate a non-parametric
       means to threshold scores generated by the decision_scores
       where outliers are set to any value below the Chauvenet's
       criterion

       Paramaters
       ----------

       method : str, optional (default='gmean')
       {'mean', 'median', 'gmean'}

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, method='gmean'):

        super(CHAU, self).__init__()
        stat = {'mean':np.mean, 'median':np.median, 'gmean':stats.gmean}
        self.method = stat[method]

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Calculate Chauvenet's criterion
        criterion = 1/(2*len(decision))

        # Get area normal to distance
        prob = erfc(np.abs(decision-self.method(decision))/decision.std())

        self.thresh_ = 1-criterion

        return 1-cut(prob, criterion)




import numpy as np
import scipy.stats as stats
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class REGR(BaseThresholder):
    """REGR class for Regression based thresholder.

       Use the regression to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond the y-intercept value of the linear fit
       
       Paramaters
       ----------

       method : str, optional (default='siegel')
       {'siegel', 'theil'}

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, method='siegel'):

        super(REGR, self).__init__()
        self.method = method

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Create a normal distribution and normalize
        norm = np.random.normal(loc=0.0, scale=1.0, size=decision.shape)
        norm = normalize(norm)

        # Set limit to the y-intercept 
        try:
            if self.method=='siegel':
                res = stats.siegelslopes(norm, decision)
            elif self.method=='theil':
                res = stats.theilslopes(norm, decision)
        except MemoryError:
            res = [0.0,1.0]
            
        limit = res[1]

        self.thresh_ = limit

        return cut(decision, limit)



import numpy as np
import scipy.stats as stats
from sklearn.metrics import auc
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde

from .iqr import IQR
from .mad import MAD
from .fwfm import FWFM
from .yj import YJ
from .kmeans import KMEANS
from .zscore import ZSCORE
from .aucp import AUCP
from .qmcd import QMCD
from .fgd import FGD
from .dsn import DSN
from .clf import CLF
from .gf import GF
from .wind import WIND
from .eb import EB
from .regr import REGR
from .boot import BOOT
from .mcst import MCST
from .hist import HIST
from .moll import MOLL
from .chau import CHAU
from .gesd import GESD
from .mtt import MTT
from .shift import SHIFT


class ALL(BaseThresholder):
    """ALL class for Combined thresholder.

       Use the multiple thresholders as a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond the (mean, median, or gmean) of the
       contamination from all the combined thresholders.
       
       Paramaters
       ----------

       thresholders : list, optional (default='all')
            list of instantiated thresholders, e.g. [DSN()]
       
       max_contam : float, optional (default=0.5)

       method : str, optional (default='mean')
           statistic to apply to contamination levels
           {'mean', 'median', 'gmean'}

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, thresholders='all', max_contam=0.5, method='mean'):

        self.thresholders = thresholders
        self.max_contam = max_contam
        stat = {'mean':np.mean, 'median':np.median, 'gmean':stats.gmean}
        self.method = stat[method]

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.

        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = np.sort(normalize(decision))

        # Initialize thresholders
        if self.thresholders=='all':
            self.thresholders = [IQR(), MAD(), FWFM(), YJ(), KMEANS(),
                                 ZSCORE(), AUCP(), QMCD(), FGD(), DSN(),
                                 CLF(), GF(), WIND(), EB(), REGR(), BOOT(),
                                 MCST(), HIST(), MOLL(), CHAU(), GESD(),
                                 MTT(), SHIFT()]

        # Apply each thresholder
        contam = []
        
        for thresholder in self.thresholders:
            
            labels = thresholder.eval(decision)
            outlier_ratio = np.sum(labels)/len(labels)
            
            if outlier_ratio<self.max_contam:
            
                contam.append(outlier_ratio)

        # Get [mean, median, or gmean] of inliers
        inlier_ratio = 1-self.method(np.array(decision))
        limit = decision[int(len(decision)*inlier_ratio)]
        
        self.thresh_ = limit
        
        return cut(decision, limit)



import numpy as np
import scipy.stats as stats
import scipy.special as special
import scipy.spatial.distance as distance
from itertools import combinations
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde

def JS_metric(val_data, val_norm):
    """Calculate the Jensen-Shannon distance"""

    return 1-distance.jensenshannon(val_data, val_norm)

def WS_metric(val_data, val_norm):
    """Calculate the Wasserstein or Earth Movers distance"""

    return stats.wasserstein_distance(val_data, val_norm)

def ENG_metric(val_data, val_norm):
    """Calculate the Energy distance"""

    return stats.energy_distance(val_data, val_norm)

def BHT_metric(val_data, val_norm):
    """Calculate the Bhattacharyya distance"""

    val_data = val_data/np.sum(val_data)
    val_norm = val_norm/np.sum(val_norm)

    bht = np.sum(np.sqrt(val_data*val_norm))
    return np.log1p(bht)

def HLL_metric(val_data, val_norm):
    """Calculate the Hellinger distance"""

    val_data = val_data/np.sum(val_data)
    val_norm = val_norm/np.sum(val_norm)

    return (1-distance.euclidean(np.sqrt(val_data), np.sqrt(val_norm))
                     /np.sqrt(2))

def HI_metric(val_data, val_norm):
    """Calculate the Histogram intersection distance"""

    val_data = val_data/np.sum(val_data)
    val_norm = val_norm/np.sum(val_norm)
    
    return np.sum(np.minimum(val_data,val_norm))

def LK_metric(val_data, val_norm):
    """Calculate the Lukaszyk–Karmowski metric for normal distributions"""
            
    # Get expected values for both distributions
    rng = np.linspace(-1,1,len(val_data))
    exp_data = (rng*val_data).sum()/val_data.sum()
    exp_norm = (rng*val_norm).sum()/val_norm.sum()
            
    nu_xy = np.abs(exp_data-exp_norm)

    # STD is same for both distributions
    std = np.std(rng)

    # Get the 1-LK distance
    return 1-(nu_xy + 2*std/np.sqrt(np.pi)*np.exp(-nu_xy**2/(4*std**2))
                     - nu_xy*special.erfc(nu_xy/(2*std)))

def LP_metric(decision, norm):
    """Calculate the Levy-Prokhorov metric"""

    # Get the edges for the complete graphs of the datasets
    f1 = np.array(list(combinations(decision.tolist(),2)))
    f2 = np.array(list(combinations(norm.tolist(),2)))

    return distance.directed_hausdorff(f1,f2)[0]

def MAH_metric(val_data, val_norm):
    """Calculate the Mahalanobis distance"""

    val_data = val_data/np.sum(val_data)
    val_norm = val_norm/np.sum(val_norm)

    V = np.cov(np.array([val_data,val_norm]).T)
    IV = np.linalg.inv(V)
    return distance.mahalanobis(val_data,val_norm,IV)/len(val_data)

def TMT_metric(val_data, val_norm):
    """Calculate the Tanimoto distance"""
    
    p = np.sum(val_data)
    q = np.sum(val_norm)
    m = np.sum(np.minimum(val_data,val_norm))

    return 1-(p+q-2*m)/(p+q-m)

def RES_metric(val_data, val_norm):
    """Calculate the studentized residual distance"""

    mean_X = np.mean(val_data)
    mean_Y = np.mean(val_norm)
    n = len(val_data)
    
    diff_mean_sqr = np.dot((val_data - mean_X), (val_data - mean_X))
    beta1 = np.dot((val_data - mean_X), (val_norm - mean_Y)) / diff_mean_sqr
    beta0 = mean_Y - beta1 * mean_X
    
    y_hat = beta0 + beta1 * val_data
    residuals = val_norm - y_hat
    
    h_ii = (val_data - mean_X) ** 2 / diff_mean_sqr + (1 / n)
    Var_e = np.sqrt(np.sum((val_norm - y_hat) ** 2)/(n-2))
    
    SE_regression = Var_e*((1-h_ii) ** 0.5)
    studentized_residuals = residuals/SE_regression
    
    return np.abs(np.sum(studentized_residuals))


class DSN(BaseThresholder):
    """DSN class for Distance Shift from Normal thresholder.

       Use the distance shift from normal to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond the distance calculated by the selected
       metric
       
       Paramaters
       ----------

       metric : str, optional (default='LK')
        {'JS', 'WS', 'ENG', 'BHT', 'HLL', 'HI', 'LK', 'LP', 'MAH', 'TMT', 'RES'}
        'JS':  Jensen-Shannon distance
        'WS':  Wasserstein or Earth Movers distance
        'ENG': Energy distance
        'BHT': Bhattacharyya distance
        'HLL': Hellinger distance
        'HI':  Histogram intersection distance
        'LK':  Lukaszyk–Karmowski metric for normal distributions
        'LP':  Levy-Prokhorov metric
        'MAH': Mahalanobis distance
        'TMT': Tanimoto distance
        'RES': Studentized residual distance

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, metric='LK'):

        super(DSN, self).__init__()
        self.metric = metric
        self.metric_funcs = {'JS': JS_metric, 'WS': WS_metric,
                             'ENG': ENG_metric, 'BHT': BHT_metric,
                             'HLL': HLL_metric, 'HI': HI_metric,
                             'LK':LK_metric, 'LP': LP_metric,
                             'MAH': MAH_metric, 'TMT': TMT_metric,
                             'RES':RES_metric}

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        #Create a normal distribution and normalize
        norm = np.random.normal(loc=0.0, scale=1.0, size=decision.shape)
        norm = normalize(norm)

        if (self.metric!='LP') or (self.metric!='RES'):
            # Create a KDE of the decision scores and the normal distribution
            # Generate KDE
            val_data, _ = gen_kde(decision,-1,1,len(decision)*3)
            val_norm, _ = gen_kde(norm,-1,1,len(decision)*3)

            val_data = val_data/np.max(val_data)
            val_norm = val_norm/np.max(val_norm)

        
            limit = self.metric_funcs[str(self.metric)](val_data, val_norm)
            
        else:
            limit = self.metric_funcs[str(self.metric)](np.sort(decision), norm)

        self.thresh_ = limit

        return cut(decision, limit)



import numpy as np
import scipy.stats as stats
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class YJ(BaseThresholder):
    """YJ class for Yeo-Johnson transformation thresholder.

       Use the Yeo-Johnson transformation to evaluate
       a non-parametric means to threshold scores generated by the
       decision_scores where outliers are set to any value beyond the
       max value in the YJ transformed data

       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(YJ, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Generate KDE
        val, _ = gen_kde(decision,0,1,len(decision)*3)

        # Use Yeo-Johnson transformation to reshape distribution
        # iterate to get average transformation
        mean_s = np.zeros(len(val))
        for i in range(50):
            scores = stats.yeojohnson(val)[0]
            mean_s += scores
        mean_s = mean_s/50

        # Set limit to the max value from the transformation
        limit = np.max(mean_s)

        self.thresh_ = limit

        return cut(decision, limit)




import numpy as np
from sklearn.cluster import MeanShift, estimate_bandwidth
from scipy.spatial.distance import cityblock
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class SHIFT(BaseThresholder):
    """SHIFT class for Mean Shift clustering thresholder.

       Use the mean shift clustering to evaluate a non-parametric means to
       threshold scores generated by the decision_scores where outliers
       are set to any value beyond the minimum labeled outlier

       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(SHIFT, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Get quantile value for bandwidth estimation
        q = cityblock(decision, np.sort(decision))/np.sum(decision)
        
        if q>1.0: q=1.0

        decision = decision.reshape(-1,1)
        
        # Estimate bandwidth
        try:
            bw = estimate_bandwidth(decision, quantile=q)
        except ValueError:
            bw = estimate_bandwidth(decision)

        # Apply mean shift and get labels
        if bw!=0:
            ms = MeanShift(bandwidth=bw, cluster_all=False, max_iter=500)
        else:
            ms = MeanShift(bandwidth=None, cluster_all=False, max_iter=500)

        ms.fit(decision)
        labels = ms.labels_

        # Set limit to smallest outlier label
        idx = np.where(labels==-1)

        if len(idx[0])!=0:  
            limit = np.min(decision[idx])
        else:
            limit = 1.1

        self.thresh_ = limit 

        return cut(np.squeeze(decision), limit)




import numpy as np
from scipy import ndimage as ndi
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde

#https://github.com/scikit-image/scikit-image/blob/v0.19.2/skimage/filters/thresholding.py

def histogram(decision, nbins):
    """Generate histograms and get bin ceneters"""

    counts, bin_edges = np.histogram(decision, bins=nbins, range=(0,1))
    bin_centers = (bin_edges[:-1] + bin_edges[1:])/2.

    return bin_centers, counts

def find_local_maxima_idx(hist):
    """Find the local maxima in histogram"""

    maximum_idxs = list()
    direction = 1

    for i in range(hist.shape[0] - 1):
        if direction > 0:
            if hist[i + 1] < hist[i]:
                direction = -1
                maximum_idxs.append(i)
        else:
            if hist[i + 1] > hist[i]:
                direction = 1

    return maximum_idxs


def OTSU_thres(bin_centers, counts):
    """Otsu's method for histogram based thresholding"""

    counts = counts.astype(float)

    #class probabilities for all possible thresholds
    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]
        
    # class means for all possible thresholds
    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:])**2

    idx = np.argmax(variance12)

    return bin_centers[idx]

def YEN_thres(bin_centers, counts):
    """Yen's method for histogram based thresholding"""

    # Calculate probability mass function
    pmf = counts.astype(np.float32) / counts.sum()
    P1 = np.cumsum(pmf) 
    P1_sq = np.cumsum(pmf ** 2)
        
    # Get cumsum calculated from end of squared array:
    P2_sq = np.cumsum(pmf[::-1] ** 2)[::-1]
        
    # Get critical value
    crit = np.log(((P1_sq[:-1] * P2_sq[1:]) ** -1) *
                    (P1[:-1] * (1.0 - P1[:-1])) ** 2)
    
    return bin_centers[crit.argmax()]

def ISODATA_thres(bin_centers, counts):
    """ISODATA method for histogram based thresholding"""
    
    counts = counts.astype(np.float32)

    # csuml and csumh contain the count of pixels in that bin or lower, and
    # in all bins strictly higher than that bin, respectively
    csuml = np.cumsum(counts)
    csumh = csuml[-1] - csuml

    # intensity_sum contains the total score intensity from each bin
    intensity_sum = counts * bin_centers

    # Get the lower and higher average value of all scoresin that bin or lower, and
    # in all bins strictly higher than that bin, respectively.
    csum_intensity = np.cumsum(intensity_sum)
    lower = csum_intensity[:-1] / csuml[:-1]
    higher = (csum_intensity[-1] - csum_intensity[:-1]) / csumh[:-1]

    # Find threshold values that meet the criterion t = (l + m)/2
    all_mean = (lower + higher) / 2.0
    bin_width = bin_centers[1] - bin_centers[0]

    # Look only at thresholds that are below the actual all_mean value
    distances = all_mean - bin_centers[:-1]
    
    return  bin_centers[:-1][(distances >= 0) & (distances < bin_width)][0]

def LI_thres(decision, bin_centers, counts):
    """Li's iterative Minimum Cross Entropy method for histogram
       based thresholing"""

    counts = counts.astype(float)

    tolerance = np.min(np.diff(np.unique(decision)))/2
    t_next = np.mean(decision) # initial new guess for iteration
    t_curr = -2 * tolerance # initial old guess for iteration

    
    # Iterate until the new and old thresholds difference
    # is less than the tolerance
    while abs(t_next - t_curr)>tolerance:
        
        t_curr = t_next
        outlier = bin_centers>t_curr
        inlier = ~outlier

        mean_out = np.average(bin_centers[outlier],
                                weights=counts[outlier])
        
        mean_in = np.average(bin_centers[inlier],
                                weights=counts[inlier])

        t_next = ((mean_in - mean_out)
                    / (np.log(mean_in) - np.log(mean_out)))
        
    return t_next

def Minimum_thres(bin_centers, counts):
    """Minimum method for histogram based thresholding"""

    smooth_hist = counts.astype(np.float64, copy=False)

    for counter in range(10000):
        smooth_hist = ndi.uniform_filter1d(smooth_hist, 3)
        maximum_idxs = find_local_maxima_idx(smooth_hist)
        if len(maximum_idxs) < 3:
            break

    # Find lowest point between the maxima
    threshold_idx = np.argmin(smooth_hist[maximum_idxs[0]:maximum_idxs[1] + 1])

    return bin_centers[maximum_idxs[0] + threshold_idx]

def Triangle_thres(bin_centers, counts):
    """Triangle algorithm for histogram based thresholding"""

    nbins = len(counts)

    # Find peak, lowest and highest score levels.
    arg_peak_height = np.argmax(counts)
    peak_height = counts[arg_peak_height]
    arg_low_level, arg_high_level = np.where(counts > 0)[0][[0, -1]]

    # Flip is True if left tail is shorter.
    flip = arg_peak_height - arg_low_level < arg_high_level - arg_peak_height
    if flip:
        counts = counts[::-1]
        arg_low_level = nbins - arg_high_level - 1
        arg_peak_height = nbins - arg_peak_height - 1

    # Set up the coordinate system.
    width = arg_peak_height - arg_low_level
    x1 = np.arange(width)
    y1 = counts[x1 + arg_low_level]

    # Normalize.
    norm = np.sqrt(peak_height**2 + width**2)
    peak_height /= norm
    width /= norm

    # Maximize the length.
    length = peak_height * x1 - width * y1
    arg_level = np.argmax(length) + arg_low_level

    if flip:
        arg_level = nbins - arg_level - 1

    return bin_centers[arg_level]
    
   

class HIST(BaseThresholder):
    """HIST class for Histogram based thresholders.

       Use histograms methods as described in scikit-image.filters to
       evaluate a non-parametric means to threshold scores generated by
       the decision_scores where outliers are set by histogram generated
       thresholds depending on the selected methods. 
       
       Paramaters
       ----------
       nbins : number of bins to use in the hostogram
               int, optional (default=None)
               
       method : str, optional (default='otsu')
               {'otsu', 'yen', 'isodata', 'li', 'minimum', 'triangle'}

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self, method='otsu', nbins=None):

        super(HIST, self).__init__()
        self.nbins = nbins
        self.method = method
        self.method_funcs = {'otsu': OTSU_thres, 'yen': YEN_thres,
                             'isodata': ISODATA_thres, 'li': LI_thres,
                             'minimum': Minimum_thres,
                             'triangle': Triangle_thres}

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        #  Set adaptive default if bins are None
        if self.nbins is None:
            self.nbins = int(len(decision)*0.7)

        # Generate histogram 
        bin_centers, counts = histogram(decision, self.nbins)

        # Threshold histogram
        if self.method!='li':
            threshold = self.method_funcs[str(self.method)](bin_centers, counts)

        else:
            threshold = self.method_funcs[str(self.method)](decision, bin_centers, counts)

        # Evaluate scores and create labels
        outliers = np.where(decision>threshold)
        scores = np.zeros(len(decision), dtype=int)
        scores[outliers] = 1

        self.thresh_ = threshold
        
        return  scores



import numpy as np
import scipy.stats as stats
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut


class MCST(BaseThresholder):
    """MCST class for Monte Carlo Shapiro Tests thresholder.

       Use uniform random sampling and statstical testing to evaluate a
       non-parametric means to threshold scores generated by the decision_scores
       where outliers are set to any value beyond the minimum value left after
       iterative Shapiro-Wilk tests have occured. Note** accuracy decreases with
       array size. For good results the should be array<1000. However still this
       threshold method may fail at any array size.
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        pass

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Get Baseline Shapiro-Wilk test p value
        p_std = stats.shapiro(decision).pvalue

        # Create random dataset to insert and test p values
        rnd = np.random.uniform(0,1,len(decision))
        povr = []

        # Iterate and add a new random variable
        # Perform a Shapiro-Wilk test and see if the new
        # distribution has a lower or higher value
        # If higher record these potential outlier values
        for i in range(len(rnd)):

            arr = np.append(decision, rnd[i])
            p_check = stats.shapiro(arr).pvalue

            if p_check>p_std:

                p_std = p_check
                povr.append(rnd[i])       

        # Set the limit to the minimum found outlier
        if len(povr)>0:
            limit = np.min(povr)
        else:
            limit = 1.0

        self.thresh_ = limit

        return cut(decision, limit)



import numpy as np
from scipy.signal import find_peaks, peak_widths
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class FWFM(BaseThresholder):
    """FWFM class for Full Width at Full Minimum thresholder.

       Use the full width at full minimum (aka base width) to evaluate
       a non-parametric means to threshold scores generated by the
       decision_scores where outliers are set to any value beyond the base
       width

       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(FWFM, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Generate KDE
        val, _ = gen_kde(decision,-1,1,len(decision)*3)
        val = normalize(val)

        # Find the greatest peak of the KDE 
        peaks, _ = find_peaks(val, prominence=0.75)

        # Find the base width of the peak
        base_width = peak_widths(val, peaks, rel_height=0.99)[0]

        # Normalize and set limit
        limit = base_width/len(val)
        self.thresh_ = limit

        return cut(decision, limit)




import numpy as np
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde


class EB(BaseThresholder):
    """EB class for Elliptical Boundary thresholder.

       Use pseudo-random elliptical boundaries to evaluate a non-parametric means
       to threshold scores generated by the decision_scores where outliers
       are set to any value beyond a pseudo-random elliptical boundary set
       between inliers and outliers 
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        pass

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Generate KDE
        val, dat_range = gen_kde(decision,0,1,len(decision)*3)

        val = val/np.max(val)

        # Generate random set of eccentricities to test
        rnd = np.random.uniform(0,1,5000)

        # Create pseudo-random elliptical boundaries using each eccentricity
        # and compute the inlier/outlier labels
        counts = [] # Get number of inliers
        for i in range(5000):
            e = rnd[i]
            a = 1/(1+e)
            lb = cut(decision, a*(1-e))
            counts.append(len(decision)-np.sum(lb))

        # Calculate the median count of expected inliers
        med = np.round(np.median(counts))
        ec = np.linspace(0,1,5000)

        # Randomly find eccentricity that generates
        # the closest value to the median inliers
        close = 0
        for i in range(5000):
            e = ec[i]
            a = 1/(1+e)
            lb = cut(decision, a*(1-e))
            count = len(decision)-np.sum(lb)
            if abs(med-count)<abs(med-close):
                close = count
                limit = a*(1-e)

        self.thresh_ = limit
        
        return cut(decision, limit)



import numpy as np
from scipy import integrate
from sklearn.utils import check_array
from .base import BaseThresholder
from .thresh_utility import normalize, cut, gen_kde

def dtheta(x,y,dx,dy,r2):
    """Calculate dtheta for the intergrand"""
    return (x*dy - y*dx)/r2


class WIND(BaseThresholder):
    """WIND class for topological Winding number thresholder.

       Use the topological winding number (with respect to the origin) to
       evaluate a non-parametric means to threshold scores generated by
       the decision_scores where outliers are set to any value beyond the
       mean intersection point calculated from the winding number
       
       Paramaters
       ----------

       Attributes
       ----------

       eval_: numpy array of binary labels of the training data. 0 stands
           for inliers and 1 for outliers/anomalies.

    """

    def __init__(self):

        super(WIND, self).__init__()

    def eval(self, decision):
        """Outlier/inlier evaluation process for decision scores.

        Parameters
        ----------
        decision : np.array or list of shape (n_samples)
                   which are the decision scores from a
                   outlier detection.
        
        Returns
        -------
        outlier_labels : numpy array of shape (n_samples,)
            For each observation, tells whether or not
            it should be considered as an outlier according to the
            fitted model. 0 stands for inliers and 1 for outliers.
        """

        decision = check_array(decision, ensure_2d=False)

        decision = normalize(decision)

        # Create a normal distribution and normalize
        norm = np.random.normal(loc=0.0, scale=1.0, size=decision.shape)
        norm = normalize(norm)

        # Create a KDE of the labels and the normal distribution
        # Generate KDE
        val_data, dat_range = gen_kde(decision,0,1,len(decision)*3)
        val_norm, _ = gen_kde(norm,0,1,len(decision)*3)

        # Get the rsquared value
        r2 = val_data**2 + val_norm**2

        val_data = val_data/np.max(val_data)
        val_norm = val_norm/np.max(val_norm)

        # Find the first derivatives of the decision and norm kdes
        # with respect to the decision scores
        deriv_data = np.gradient(val_data, dat_range[1]-dat_range[0])
        deriv_norm = np.gradient(val_norm, dat_range[1]-dat_range[0])

        # Compute intergrand
        integrand = dtheta(val_data,val_norm,deriv_data,deriv_norm,r2)

        # Intergrate to find winding numbers mean intersection point
        limit = integrate.simpson(integrand)/np.sum((val_data+val_norm)/2)

        self.thresh_ = limit 
          
        return cut(decision, limit)
